#ifndef CGTKPKGLIST_H
#define CGTKPKGLIST_H
#include <sstream>
#include <gtkmm.h>
#include <map>
#include "CVersionCompare.h"
#include "common_fun.h"

// A widget to hold the list of packages for zendo.
// The actual package list is held in an STL map.
// It also adds a popup menu for selecting the actions to perform
// on the packages too, and it supports reading files from the hard drive
// to parse the contents of them and generate a package list.
// See COPYING for license information.

enum pkgfilter_e {
	EALL,
	ENEW,
	EINSTALLED,
	EUPGRADES,
	EBOTHGRADES,
	EBLACKLISTED
};

class CGtkPkgList : public Gtk::ScrolledWindow
{
public:
	CGtkPkgList();
	void		UpdateDisplay(Glib::ustring cat, pkgfilter_e filter, Glib::ustring searchtext = "", bool searchdesc = false);
	Gtk::TreeView*	GetTreeView();
	Glib::ustring	GetCurrentPackageName();
	Glib::ustring	GetMirrorCount();
	Glib::ustring	GetInstalledCount();
	void		MarkAll();
	void		MarkNone();

	packagemap_t	Packages;
	Glib::ustring	Blacklist;
protected:
	bool		categoryMatch(Glib::ustring& location, Glib::ustring& category);
	void		on_checkbutton(Glib::ustring const& string_path);
	void		on_buttonpress(GdkEventButton* eb);
	void		on_menu_install();
	void		on_menu_remove();
	void		on_menu_blacklist();
	void		on_menu_unblacklist();
	void		on_treeview_doubleclicked(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column);
	bool		toggleStatus();

	Gtk::TreeView						_TreeView;
	Gtk::TreeModel::ColumnRecord				_ColumnRecord;
	Glib::RefPtr<Gtk::ListStore>				_RefModel;
	Gtk::TreeModelColumn<bool>				_ColSelect;
	Gtk::TreeModelColumn<Glib::ustring>			_ColName;
	Gtk::TreeModelColumn< Glib::RefPtr<Gdk::Pixbuf> >	_ColPixmap;
	Gtk::TreeModelColumn<Glib::ustring>			_ColMirrorVersion;
	Gtk::TreeModelColumn<Glib::ustring>			_ColInstalledVersion;
	Gtk::TreeModelColumn<guint32>				_ColSize;
	Gtk::TreeModelColumn<Glib::ustring>			_ColColor;	// Not displayed

	Gtk::Menu	_Menu;
	CVersionCompare	_Vcmp;
};

CGtkPkgList::CGtkPkgList()
{
	set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	add(_TreeView);
	_ColumnRecord.add(_ColSelect);
	_ColumnRecord.add(_ColName);
	_ColumnRecord.add(_ColPixmap);
	_ColumnRecord.add(_ColMirrorVersion);
	_ColumnRecord.add(_ColInstalledVersion);
	_ColumnRecord.add(_ColSize);
	_ColumnRecord.add(_ColColor);	// This column is not displayed. For educational purposes only. JK ;)

	_RefModel = Gtk::ListStore::create(_ColumnRecord);
	_TreeView.set_model(_RefModel);
	// Double-click and right-click callbacks.
	_TreeView.signal_row_activated().connect(sigc::mem_fun(*this, &CGtkPkgList::on_treeview_doubleclicked));
	_TreeView.signal_button_press_event().connect_notify(sigc::mem_fun(*this, &CGtkPkgList::on_buttonpress));

	// Set up the checkbutton column to have color and a callback.
	Gtk::CellRendererToggle*	_cell = new Gtk::CellRendererToggle;
	Gtk::TreeViewColumn*		_pcolumn = _TreeView.get_column(_TreeView.append_column("", *_cell) - 1);
	_pcolumn->add_attribute(*_cell, "active", _ColSelect);
	_pcolumn->add_attribute(_cell->property_cell_background(), _ColColor);
	_cell->signal_toggled().connect(sigc::mem_fun(*this, &CGtkPkgList::on_checkbutton));

	// Add the rest of the columns.
	_TreeView.append_column("Package Name", _ColName);
	_TreeView.append_column("", _ColPixmap);
	_TreeView.append_column("Mirror Version", _ColMirrorVersion);
	_TreeView.append_column("Installed Version", _ColInstalledVersion);
	_TreeView.append_column("Size (KB)", _ColSize);

	// Enable rudimentary searching (Find Ahead) and sorting.
	_TreeView.set_search_column(1);
	_TreeView.get_column(0)->set_sort_column(_ColSelect);
	_TreeView.get_column(1)->set_sort_column(_ColName);
	_TreeView.get_column(2)->set_sort_column(_ColPixmap);
	_TreeView.get_column(3)->set_sort_column(_ColMirrorVersion);
	_TreeView.get_column(4)->set_sort_column(_ColInstalledVersion);
	_TreeView.get_column(5)->set_sort_column(_ColSize);

	// Create the menu for selecting actions to perform on packages.
	Gtk::Menu::MenuList&	_menulist = _Menu.items();
	_menulist.push_back(Gtk::Menu_Helpers::MenuElem("_Install/Update",
		sigc::mem_fun(*this, &CGtkPkgList::on_menu_install)));
	_menulist.push_back(Gtk::Menu_Helpers::MenuElem("_Remove",
		sigc::mem_fun(*this, &CGtkPkgList::on_menu_remove)));
	_menulist.push_back(Gtk::Menu_Helpers::MenuElem("_Blacklist",
		sigc::mem_fun(*this, &CGtkPkgList::on_menu_blacklist)));
	_menulist.push_back(Gtk::Menu_Helpers::MenuElem("_Unblacklist",
		sigc::mem_fun(*this, &CGtkPkgList::on_menu_unblacklist)));
	_Menu.accelerate(*this);
}

void CGtkPkgList::UpdateDisplay(Glib::ustring cat, pkgfilter_e filter, Glib::ustring searchtext, bool searchdesc)
{	// Takes a packagemap_t and displays its contents in this widget
	Gtk::TreeModel::Row	_row;
	_RefModel->clear();
	for ( packagemap_t::iterator _iter = Packages.begin(); _iter != Packages.end(); ++_iter ) {
		if ( !categoryMatch(_iter->second.location, cat) ) continue;
		if ( filter == ENEW && _iter->second.installedversion.length() ) continue;
		if ( filter == EINSTALLED && !_iter->second.installedversion.length() ) continue;
		if ( filter == EUPGRADES && !_Vcmp.Compare(_iter->second.mirrorversion, _iter->second.installedversion) ) continue;
		if ( filter == EBOTHGRADES && (_iter->second.mirrorversion == _iter->second.installedversion ||
			_iter->second.mirrorversion.empty() || _iter->second.installedversion.empty()) ) continue;
		if ( filter == EBLACKLISTED && _iter->second.color != COLOR_BLACKLIST ) continue;
		if ( searchtext.length() ) {	// Search for "searchtext" in the pkg desc.
			bool	_found = false;
			for ( gint16 i = 0; i < _iter->second.description.size(); i++ ) {
				if ( _found || !searchdesc ) break;
				if ( (gint16)_iter->second.description[i].casefold().find(searchtext.casefold()) >= 0 )
					_found = true;
			}
			if ( (gint16)_iter->second.name.casefold().find(searchtext.casefold()) >= 0 )
				_found = true;	// Search in the pkg names.
			if ( !_found ) continue;
		}
		_row = *(_RefModel->append());
		if ( _iter->second.installedversion.length() )
			_row[_ColSelect] = true;
		else	_row[_ColSelect] = false;
		_row[_ColName] = _iter->second.name;
		_row[_ColMirrorVersion] = _iter->second.mirrorversion;
		_row[_ColInstalledVersion] = _iter->second.installedversion;
		_row[_ColSize] = _iter->second.packedsize;
		_row[_ColColor] = _iter->second.color;

		// Show some pixmaps for upgrades and downgrades.
		if ( _iter->second.installedversion.length() && _iter->second.mirrorversion.length() ) {
			if ( _Vcmp.Compare(_iter->second.installedversion, _iter->second.mirrorversion) )
				_row[_ColPixmap] = Gdk::Pixbuf::create_from_file(PIXMAP("version_downgrade.png"));
			else if ( _Vcmp.Compare(_iter->second.mirrorversion, _iter->second.installedversion) )
				_row[_ColPixmap] = Gdk::Pixbuf::create_from_file(PIXMAP("version_upgrade.png"));
			else	_row[_ColPixmap] = Gdk::Pixbuf::create_from_file(PIXMAP("version_other.png"));
		} else if ( _iter->second.installedversion.length() )
			_row[_ColPixmap] = Gdk::Pixbuf::create_from_file(PIXMAP("version_other.png"));
		else 	_row[_ColPixmap] = Gdk::Pixbuf::create_from_file(PIXMAP("version_notinstalled.png"));
	}
}

Gtk::TreeView* CGtkPkgList::GetTreeView()
{
	return &_TreeView;
}

Glib::ustring CGtkPkgList::GetCurrentPackageName()
{	// Returns the name of the package currently selected.
	Gtk::TreeModel::iterator	_pkgiter = _TreeView.get_selection()->get_selected();
	if ( _pkgiter ) return (*_pkgiter)[_ColName];
	return "";
}

Glib::ustring CGtkPkgList::GetMirrorCount()
{	// Returns the number of packages on the mirror.
	int			_count = 0;
	std::stringstream	_ss;
	Glib::ustring		_ret;
	for ( packagemap_t::iterator _iter = Packages.begin(); _iter != Packages.end(); ++_iter )
		if ( _iter->second.mirrorversion.length() ) _count++;
	_ss << _count;
	_ss >> _ret;
	return _ret;
}

Glib::ustring CGtkPkgList::GetInstalledCount()
{	// Returns the number of packages currently installed.
	int			_count = 0;
	std::stringstream	_ss;
	Glib::ustring		_ret;
	for ( packagemap_t::iterator _iter = Packages.begin(); _iter != Packages.end(); ++_iter )
		if ( _iter->second.installedversion.length() ) _count++;
	_ss << _count;
	_ss >> _ret;
	return _ret;
}

void CGtkPkgList::MarkAll()
{	// Iterate through and select all the packages currently being viewed.
	Gtk::TreeNodeChildren	_children = _RefModel->children();
	for ( Gtk::TreeNodeChildren::iterator _iter = _children.begin(); _iter != _children.end(); _iter++ ) {
		if ( (*_iter)[_ColColor] != COLOR_BLACKLIST ) {
			(*_iter)[_ColColor] = COLOR_INSTALL;
			Packages[(*_iter)[_ColName]].color = COLOR_INSTALL;
		}
	}
}

void CGtkPkgList::MarkNone()
{
	Gtk::TreeNodeChildren	_children = _RefModel->children();
	for ( Gtk::TreeNodeChildren::iterator _iter = _children.begin(); _iter != _children.end(); _iter++ ) {
		if ( (*_iter)[_ColColor] != COLOR_BLACKLIST ) {
			(*_iter)[_ColColor] = COLOR_NEUTRAL;
			Packages[(*_iter)[_ColName]].color = COLOR_NEUTRAL;
		}
	}
}

bool CGtkPkgList::categoryMatch(Glib::ustring& location, Glib::ustring& category)
{	// Returns true if the package location matches the category we are viewing
	if ( category == "All" )
		return true;
	else if ( location == category )	// true if we are using mirror categories
		return true;
	else if ( (location == "xap" || location == "extra/xap") && (category == "Applications") )
		return true;
	else if ( (location == "ap" || location == "extra/ap") && (category == "Applications (CLI)") )
		return true;
	else if ( (location == "a" || location == "extra/a") && (category == "Base System") )
		return true;
	else if ( (location == "x" || location == "extra/x") && (category == "Base (X Windows)") )
		return true;
	else if ( (location == "e" || location == "extra/e") && (category == "Desktop (Enlightenment)") )
		return true;
	else if ( (location == "g" || location == "extra/g") && (category == "Desktop (Gnome)") )
		return true;
	else if ( (location == "kde" || location == "extra/kde") && (category == "Desktop (KDE)") )
		return true;
	else if ( (location == "games" || location == "extra/games") && (category == "Games") )
		return true;
	else if ( (location == "locale" || location == "extra/locale") && (category == "International Support") )
		return true;
	else if ( (location == "l" || location == "extra/l") && (category == "Libraries") )
		return true;
	else if ( !location.find("local://") && (category == "Local Mirror") )
		return true;
	else if ( (location == "f" || location == "extra/f") && (category == "Misc. Documentation") )
		return true;
	else if ( (location == "n" || location == "extra/n") && (category == "Network") )
		return true;
	else if ( (location == "s" || location == "extra/s") && (category == "Network (Web Server)") )
		return true;
	else if ( (location == "d" || location == "extra/d") && (category == "Programming Tools") )
		return true;
	else if ( (location == "restricted") && (category == "Restricted") )
		return true;
	else if ( (location == "tcl" || location == "extra/tcl") && (category == "Tcl/tk") )
		return true;
	else if ( (location == "t" || location == "extra/t") && (category == "teTeX") )
		return true;
	return false;
}

void CGtkPkgList::on_checkbutton(Glib::ustring const& string_path)
{	// Show context menu when the user clicks the checkbutton.
	Glib::signal_timeout().connect(sigc::mem_fun(*this, &CGtkPkgList::toggleStatus), 5);
}

void CGtkPkgList::on_buttonpress(GdkEventButton* eb)
{	// If it's a right click, show the context menu.
	if ( !_TreeView.get_selection()->get_selected() ) return;
	if ( eb->type == GDK_BUTTON_PRESS && eb->button == 3 )
		_Menu.popup(0, gtk_get_current_event_time());
}

void CGtkPkgList::on_menu_install()
{	// Mark the package for installation, if it's not blacklisted.
	// Then unselect the package, so that it is easy to see what you've done.
	Gtk::TreeModel::Row	_row = (*_TreeView.get_selection()->get_selected());
	_TreeView.get_selection()->unselect_all();
	if ( _row[_ColColor] == COLOR_BLACKLIST || !Glib::ustring(_row[_ColMirrorVersion]).length() ) return;
	_row[_ColColor] = COLOR_INSTALL;
	Packages[_row[_ColName]].color = _row[_ColColor];
}

void CGtkPkgList::on_menu_remove()
{	// Mark the package for removal, if it's installed and not blacklisted.
	Gtk::TreeModel::Row	_row = (*_TreeView.get_selection()->get_selected());
	_TreeView.get_selection()->unselect_all();
	if ( _row[_ColColor] == COLOR_BLACKLIST || !_row[_ColSelect] ) return;
	_row[_ColColor] = COLOR_REMOVE;
	Packages[_row[_ColName]].color = _row[_ColColor];
}

void CGtkPkgList::on_menu_blacklist()
{	// Add the current package to the blacklist.
	Gtk::TreeModel::Row	_row = (*_TreeView.get_selection()->get_selected());
	_TreeView.get_selection()->unselect_all();
	_row[_ColColor] = COLOR_BLACKLIST;
	Packages[_row[_ColName]].color = _row[_ColColor];
	if ( (gint32)Blacklist.find("," + Glib::ustring(_row[_ColName]).lowercase() + ",") < 0 )
		Blacklist += Glib::ustring(_row[_ColName]).lowercase() + ",";
}

void CGtkPkgList::on_menu_unblacklist()
{	// Remove the current package from the blacklist.
	Gtk::TreeModel::Row	_row = (*_TreeView.get_selection()->get_selected());
	_TreeView.get_selection()->unselect_all();
	_row[_ColColor] = COLOR_NEUTRAL;
	Packages[_row[_ColName]].color = _row[_ColColor];
	if ( (gint32)Blacklist.find("," + Glib::ustring(_row[_ColName]).lowercase() + ",") >= 0 )
		Blacklist.erase((gint32)Blacklist.find("," + Glib::ustring(_row[_ColName]).lowercase() + ",") + 1, Glib::ustring(_row[_ColName]).length() + 1);
}

void CGtkPkgList::on_treeview_doubleclicked(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{	// User double clicked one of the packages.
	Glib::signal_timeout().connect(sigc::mem_fun(*this, &CGtkPkgList::toggleStatus), 100);
}

bool CGtkPkgList::toggleStatus()
{	// Make an guess about what action should be performed on the package.
	Gtk::TreeModel::Row	_row = *(_TreeView.get_selection()->get_selected());
	if ( !_row ) return false;	// GTK hasn't caught up yet, wait a little while longer.
	_TreeView.get_selection()->unselect_all();
	if ( _row[_ColColor] == COLOR_INSTALL || _row[_ColColor] == COLOR_REMOVE )
		_row[_ColColor] = COLOR_NEUTRAL;
	else if ( _row[_ColColor] == COLOR_NEUTRAL &&
			Glib::ustring(_row[_ColMirrorVersion]) ==
			Glib::ustring(_row[_ColInstalledVersion]) )
		_row[_ColColor] = COLOR_REMOVE;
	else if ( _row[_ColColor] == COLOR_NEUTRAL && !Glib::ustring(_row[_ColMirrorVersion]).length() )
		_row[_ColColor] = COLOR_REMOVE;
	else if ( _row[_ColColor] == COLOR_NEUTRAL && Glib::ustring(_row[_ColMirrorVersion]).length() )
		_row[_ColColor] = COLOR_INSTALL;
	Packages[_row[_ColName]].color = _row[_ColColor];
	return false;	// Don't repeat this loop.
}

#endif
